/* 
 * Copyright (C) 1999-2001 Peter T. Breuer <ptb@it.uc3m.es>
 */


#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <syslog.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mount.h> // PTB for BLKGETSIZE
#include <sys/param.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <asm/errno.h>
#include <sys/sysmacros.h> // PTB for major()

#ifndef BLKGETSIZE
/* lifted from fs.h */
#define BLKGETSIZE _IO(0x12,96) 
#endif
#ifndef BLKROGET
/* lifted from fs.h */
#define BLKROGET   _IO(0x12,94) 
#endif
#ifndef BLKGETSIZE64
/* lifted from fs.h */
#define BLKGETSIZE64 _IOR(0x12,114,u64) 
#endif
/* lifted from fs.h */
#ifndef BLKBSZGET
#define BLKBSZGET  _IOR(0x12,112,int)
#endif
/* lifted from fs.h */
#ifndef BLKBSZSET
#define BLKBSZSET  _IOW(0x12,113,int)
#endif


# include "cliserv.h"
#ifdef HAVE_LIBPUB
#include <publib.h>
#endif

# include "file.h"
# include "select.h"
# include "list.h"

#ifndef MY_NAME
#define MY_NAME "file"
#endif

#ifndef FLOPPY_MAJOR
#define FLOPPY_MAJOR 2
#endif

extern int debug_level;


   static inline s64 seek(int fildes, s64 offset, int whence)
   {
       DEBUG("seek to sector %Ld type %s\n", offset >> 9,
          whence==SEEK_CUR?"CUR":whence==SEEK_SET?"SET":"SEEK_END");

#ifdef HAVE_LSEEK64
       return lseek64(fildes, (loff_t) offset, whence);
#elif HAVE_LLSEEK
       return llseek(fildes, (loff_t)offset, whence);
#else
       if ((offset >> 31) != 0) {
         PERR("Cannot seek to more than 2^31-1 (0x%Lx) in one file!\n", offset);
         return -EINVAL;
       }
       return (s64) lseek(fildes, (off_t) offset, whence);
#endif
   }


   struct nbd_seek {
       int i;
       int fd;
       u64 a; // local offset
       int direction; // 0 for an input, 1 for an output
   };

   /*
    * Fill out an nbd_seek struct given the intended offset.
    * */
   static  int
   resolve_offset(struct nbd_file *self, u64 a, struct nbd_seek *s) {

      int i;
      if (a >= self->size)
        return -EINVAL;

      switch (self->mode) {
           u64 b;

           // PTB choose the fd to write to

           case F_LINEAR_MODE:
           for (i = 0; i < self->nfile; i++) {
             u64 size = self->sizes[i];
             if (a < size)
               break;
             a -= size;
           }
           if (i >= self->nfile) {
             return -EINVAL;
           }
           break;

           case F_MIRROR_MODE:
           i = 0;   // PTB silly, but what else can I say!
           break;

           case F_STRIPE_MODE:
           b = a % self->blksize;   // offset within block
           a/= self->blksize;       // block number
           i = a % self->nfile;     // file id holding that block number
           a/= self->nfile;         // block number within file
           a*= self->blksize;
           a+= b;             // plus offset within block
           break;

           default:
           i = -1;   // PTB to avoid compiler warnings
           break;
      }

      if (s && i >= 0 && i < self->nfile) {
        s->i = i;
        s->fd = self->exports[i];
        s->a = a;
      }
      
      return 0;
   }


   /*
    * return the fd of the component we want and also as a side effect
    * set self->lastpoint and self->lastpt[self->lasti] and
    * self->lasti. We also seek in the right file to the right point.
    * We return INVAL for a whacky seek instruction (overrange,
    * underrange).  Return BADF for a failed seek.
    *
    * if s is nonnull on entry, assume where we are has already been
    * resolved in there.
    */
   static int maybeseek(struct nbd_file * self, u64 a, struct nbd_seek *s)
   {
      int i = self->lasti; 
      struct nbd_seek s0;

      if (a >= self->size) {
        return -EINVAL;
      }

      DEBUG("@%Lu",a);

      if (self->lastpoint == a && !self->changedfd) {    // PTB fast path
        DEBUG(":%d/%Lu/%Lu",
          self->lasti,
          self->lastpt[self->lasti]/self->blksize,
          self->lastpt[self->lasti]%self->blksize);
        return self->exports[i]; 
      }

      if (!s) {
          // use our own seek struct
          s = &s0;
          if ( resolve_offset(self, a, s) < 0) {
             PERR("Cannot happen to me! Overstepped FD array\n"); 
             return -EINVAL;
          }
      } else {
          if (s->direction) {
              // output
              if ( resolve_offset(self, a, s) < 0) {
                 PERR("Cannot happen to me! Overstepped FD array\n"); 
                 return -EINVAL;
              }
          } else {
              // it's an input. believe it.
          }
      }

      i = s->i;

      if (self->lastpt[i] > 0 && self->lastpt[i] == s->a) {
                           // PTB medium fast path
          self->lasti = i;
          DEBUG(":%d/%Lu/%Lu", i, s->a/self->blksize, s->a%self->blksize);
          self->lastpoint = a;    // PTB setup fast path for next time
          return self->exports[i];
      }
                            // PTB slow path

      switch (self->mode) {

        int j;
        case F_STRIPE_MODE:
        case F_LINEAR_MODE:
          if (seek(self->exports[i], s->a, SEEK_SET) < 0) {
             DEBUG("Cannot seek locally to offset %Ld!\n", s->a);
             return -EINVAL;
          }
        break;

        case F_MIRROR_MODE:
          for (j = 0; j < self->nfile; j++) {
            if (seek(self->exports[j], a, SEEK_SET) < 0) {
               DEBUG("Cannot seek locally to offset %Ld!\n", s->a);
               return -EBADF;
            }
          }
        break;

      }
      self->lasti = i;
      self->lastpt[i] = s->a;

      DEBUG(":%d/%Lu/%Lu",i,s->a/self->blksize,s->a%self->blksize);
      self->lastpoint = a;    // PTB setup fast path for next time
      return self->exports[i];
   }

  /*
   * try to close the nonclosed parts of the device. Set F_CLOSED when done.
   * Unset F_OPEN in any case if partially done. Return 0 for full success,
   * otherwise something negative.
   */
   static int
   closefile(struct nbd_file * self) {
         int j;
         int errs = 0;
         int succ = 0;

 	 DEBUG("entered\n");
         self->flags &= ~F_OPENWANTED;    // PTB want close if anything
         self->flags |= F_CLOSEWANTED;    // PTB more closes required

         if (self->flags & F_CLOSED) {
             self->flags &= ~F_CLOSEWANTED;
 	     DEBUG("exited 0\n");
             return 0;
         }

         for (j = 0; j< self->nfile; j++) {
             int export = self->exports[j];
             char * exportname = self->names[j];
             self->lastpt[j] = 0;
             if (export < 0) {
                 DEBUG("skipped already closed %s\n", exportname);
                 succ++;
                 continue;
             }
             if (close(export) < 0) {
 	        PERR("Cannot close file %s\n", exportname);
                errs++;
             } else {
                 self->exports[j] = -1;
                 succ++;
                 DEBUG("closed %s\n", exportname);
             }
         }
         self->lastpoint = 0;

         if (succ > 0)
           self->flags &= ~F_OPEN;        // PTB not fully open

         if (errs <= 0) {
           self->flags &= ~F_CLOSEWANTED;
           self->flags |= F_CLOSED;       // PTB fully closed
 	   DEBUG("exited 0 and fully closed\n");
           return 0;
         }
         DEBUG("exited %d and partially closed\n", -errs);
         return -errs;
   }

   static int
   getsize1(struct nbd_file * self, int fd, u64 *res) {

         s64 size = -1;

         if (size < self->blksize) {

               s64 es = 0;
               loff_t off;

               DEBUG("looking for export size with seek SEEK_END\n");
               off = seek(fd, 0, SEEK_CUR); 
               if (off >= 0) {
                   es = seek(fd, 0, SEEK_END);
                   if (es != -1) {
                       if (es >= self->blksize){
                           size = es;
                       }
                       seek(fd,off,SEEK_SET);
                   }
               }
         }
         if (size < self->blksize) {

               off_t es = 0;
               struct stat stat_buf;
               int error;

               memset(&stat_buf, 0, sizeof(stat_buf));
               DEBUG("looking for export size with fstat\n");
               error = fstat(fd, &stat_buf);
               if (error < 0) {
                   // PTB if we can't stat it, it doesn't exist!
                   return -EINVAL;
               }
               es = stat_buf.st_size;
               if (es >= self->blksize) { 
                   size = es;
               }
         }
         if (size < self->blksize) {

               s64 es = 0;

               DEBUG("looking for export size with ioctl BLKGETSIZE64\n");
               if (ioctl(fd, BLKGETSIZE64, &es) >= 0) {
                   if (es >= self->blksize) {
                       size = es;
                   }
               }
         }
         if (size < self->blksize) {

               off_t es = 0;

               DEBUG("looking for export size with ioctl BLKGETSIZE\n");
               if (ioctl(fd, BLKGETSIZE, &es) >= 0) {
                   if (es >= self->blksize) {
                       size = es;
                       size <<= 9; /* assume blocksize 512 */
                   }
               }
         }
         if (self->blksize <= 0) {
             // PTB paranoia
             PERR("really need blksize (currently %d) before getting size\n",
                     self->blksize);
             return -EINVAL;
         }
         size = ( size / self->blksize ) * self->blksize;
         *res = size;
         return 0;
  }

  static int
  getsize(struct nbd_file * self, s64 *res) {
     int i = 0;
     u64 totalsize = 0;
     short unknown_size = 0;

     for (i = 0; i < self->nfile; i++) {

         u64 size = 0;
         int fd = self->exports[i];

         if (fd < 0) {
                unknown_size++;
                continue;
         }

         if (self->sizes[i] <= 0) {
                if (getsize1(self, fd, &size) < 0) {
                        unknown_size++;
                        continue;
                }
         }

#ifndef HAVE_LSEEK64
#ifndef HAVE_LLSEEK
        if ((size >> 31) != 0) {
            PERR("%s is too large to export!\n", exportnames[i]);
            return -EINVAL;
        }
#endif
#endif

         self->sizes[i] = size;
         switch (self->mode) {
           case F_LINEAR_MODE: totalsize += size;
                   break;
           case F_MIRROR_MODE: totalsize = (i>0&&totalsize>0&&size<=totalsize)
                               ? : size;
                   break;
           case F_STRIPE_MODE: totalsize = (i<=0||totalsize<=0||size<totalsize/i)
                               ? size*(i+1) : (totalsize/i)*(i+1);
                   break;
         }
     }

     if (!unknown_size) {
        *res = totalsize;
        return 0;
     }

     return -EINVAL;
  } 

#ifndef HAVE_LIBPUB
  static int
  strgsub (char *str, char *pat, char* sub, size_t max)
  {
    int instances = 0;
    int patlen = strlen(pat);
    int sublen = strlen(sub);
    int strnlen = strlen(str);
    char *s = strstr(str, pat);

    while (s) {
        // found an instance of pat
        strnlen += sublen - patlen;
        if (strnlen >= max)
            // ran out of room
            return -1;
        // move the rest of the string up in order to make room
        memmove(s+sublen, s+patlen, strnlen - ((s + sublen) - str));
        // fill in the gap with sub
        memmove(s, sub, sublen);
        instances ++;
        s = strstr(s+sublen, pat);
    }
    // ran out of pattern repetitions
    return instances;
  }

#endif

   static int nsubstr(char *str, char *pat) {
       char *s = str;
       int patlen = strlen(pat);
       int n = 0;
       if (patlen == 0)
           return 1;
       while ((s = strstr(s, pat)) != NULL) {
           s += patlen;
           n++;
       }
       return n;
   }

  static int
  rootstatblkdev(char *exportname, unsigned long *flags) {

	    // PTB block device can tell us via BLKROGET

	    int openflags = O_NONBLOCK | O_RDONLY; // just for IOCTL
	    int export;
	    int intval = -1;

#ifdef O_LARGEFILE
	    openflags |= O_LARGEFILE;
#endif
	    export = open (exportname, openflags);

	    if (export < 0) {
		PERR ("Could not open exported file %s: %m\n", exportname);
		// definitely not even openable RO async
		return -EINVAL;
	    }

	    if (ioctl (export, BLKROGET, &intval) < 0 || intval == -1) {
		PERR ("Warning, could not query %s with BLKROGET: %m\n",
		      exportname);
		PERR ("Warning, assuming %s is readwrite:\n", exportname);
		// PTB assume rw
		intval = 0;
	    }

	    close (export);

	    if (!intval) {
		// is rw
		return 0;
	    }

	    // is RO
	    MSG ("set device ro because component %s is ro\n", exportname);
	    *flags |= F_READONLY;
	    return 0;
  }

  static int
  rootstatfile(char *exportname, unsigned long *flags) {
        // PTB not a block device. Probably a file. Try real open.

	int openflags = O_RDWR;
	int export;

#ifdef O_LARGEFILE
	openflags |= O_LARGEFILE;
#endif
	export = open (exportname, openflags);

	if (export < 0) {
	        PERR ("Could not open exported file %s rw: %m\n",
		      exportname);
		// is RO
		MSG ("set device ro because component %s is ro\n",
		     exportname);
		*flags |= F_READONLY;
		return 0;
	}

	close (export);

	// PTB was root and no negative decision made. Move on.
	return 0;
  }

  /*
   * set RO bit of flag if given file is definitely readonly.
   */
  static int
  userstatfile(char *exportname, unsigned long *flags, struct stat *buf) {

    if (buf->st_mode & S_IWOTH) {
	DEBUG ("component %s is writable by everyone\n", exportname);
	return 0;
    }

    // writable by our group
    if ((buf->st_mode & S_IWGRP)
	&& buf->st_gid == getegid ()) {
	DEBUG ("component %s is writable by our group %d\n", exportname,
	       getegid ());
	return 0;
    }

    // writable by owner and owned by us
    if ((buf->st_mode & S_IWUSR)
	&& buf->st_uid == geteuid ()) {
	DEBUG ("component %s is writable by owner me %d\n", exportname,
	       geteuid ());
	return 0;
    }

    // writable by group and it's one of our groups
    if (buf->st_mode & S_IWGRP) {
	gid_t grps[NGROUPS];
        int ngrps;
        int j;

	ngrps = getgroups (NGROUPS, grps);

	for (j = 0; j < ngrps; j++) {
	    if (buf->st_gid == grps[j]) {
		DEBUG ("component %s is writable by my group %d\n",
		       exportname, grps[j]);
		break;
	    }
	}
	if (j < ngrps && ngrps > 0) {
	    DEBUG ("component %s is writable by my group %d\n",
		   exportname, grps[j]);
	    return 0;
	}
    }

    return 0;

  }

  /*
   * set RO bit of flag if given block device is definitely readonly.
   */
  static int
  userstatblkdev(char *exportname, unsigned long *flags) {

        // Block device

	int openflags = O_NONBLOCK | O_RDONLY;
	int export;
	int intval = -1;

#ifdef O_LARGEFILE
	openflags |= O_LARGEFILE;
#endif
	export = open (exportname, openflags);	// for an ioctl

	if (export < 0) {
	    PERR ("Could not open exported file %s: %m\n", exportname);
	    return -EINVAL;
	}

	if (ioctl (export, BLKROGET, &intval) < 0 || intval == -1) {
	    DEBUG ("Could not query exported file %s: %m\n", exportname);
            // assume r/w
            intval = 0;
	}

	close (export);

        if (intval)
            *flags |= F_READONLY;

        return 0;
  }

  /*
   * set RO bit of flag if given file or blkdev is definitely readonly.
   */
  static int
  rostat1(char *exportname, unsigned long *flags) {

      // really check!

	int openflags = O_NONBLOCK | O_RDONLY;
        int export;

#ifdef O_LARGEFILE
	openflags |= O_LARGEFILE;
#endif
	export = open (exportname, openflags);
	if (export < 0) {
	    DEBUG ("Could not open exported file %s rw: %m\n", exportname);
	    return -EINVAL;
	}
	close (export);

        // PTB we know it opens RO, does it open RW?
	openflags = O_NONBLOCK | O_RDWR;
#ifdef O_LARGEFILE
	openflags |= O_LARGEFILE;
#endif

	export = open (exportname, openflags);
	if (export < 0) {
	    // is RO
	    MSG ("set device ro because component %s is ro\n", exportname);
	    *flags |= F_READONLY;
	    return 0;
	}
	close (export);

        return 0;
  }


  /*
   * set RO bit of flag if given device or file is definitely readonly.
   */
  static int
  userstat1(char *exportname, unsigned long *flags, struct stat *buf) {
    // PTB ordinary people have to look at perms
    int errs;

    if (S_ISBLK (buf->st_mode)) {
        errs = userstatblkdev(exportname, flags);
    } else {
        errs = userstatfile(exportname, flags, buf);
    }
    return errs;
  }

  /*
   * set RO bit of flag if given device or file is definitely readonly.
   */
  static int
  rootstat1(char *exportname, unsigned long *flags, struct stat *buf) {

        int errs;

	if (S_ISBLK (buf->st_mode)) {
            errs = rootstatblkdev(exportname, flags);
        } else {
            errs = rootstatfile(exportname, flags);
        }
        return errs;
  }
 
  /*
   * set RO bit of flag if given device or file is definitely readonly.
   * Also set other flags as necessary
   */
  static int
  statfile1 (char *exportname, unsigned long *flags)
  {

    struct stat buf;
    int errs;

    if (stat (exportname, &buf) < 0) {
	PERR ("Could not stat exported file %s: %m\n", exportname);
	return -EINVAL;
    }

    if (S_ISBLK (buf.st_mode) && major (buf.st_rdev) == FLOPPY_MAJOR) {
	if (!(*flags & F_ASYNC) && !(*flags & F_SYNC)) {
	    // is floppy, make us sync
	    *flags |= F_SYNC;
	    MSG ("set sync flags on %s because it's a floppy\n", exportname);
        }
	if (!(*flags & F_REMOVABLE)) {
	    // is floppy, make us removable
	    *flags |= F_REMOVABLE;
	    MSG ("set removable flags on %s because it's a floppy\n", exportname);
        }
    }

    // PTB ro checks

    if (geteuid () == 0)  {
        errs = rootstat1(exportname, flags, &buf);
    } else {
        errs = userstat1(exportname, flags, &buf);
    }

    // PTB final ro check
    //
    if (errs >= 0 && ! (*flags & F_READONLY)) {
        errs = rostat1(exportname, flags);
    }

    return errs;
  }

  /*
   * Replace %s in filename by client address. The filename has
   * already been expanded to become absolute by this point (not
   * that that matters for the purposes of this routine)
   */
  static char *
  expandname (char *exportname, unsigned long caddr)
  {
    char *s;
    int maxlen;
    char s_caddr[16];
    int n;

    if (caddr == 0) {
	// PTB we need to substitute and we can't.
	PERR ("no information yet to expand filename %s: %m\n", exportname);
	return NULL;
    }

    n = nsubstr(exportname, "%s");
    maxlen = strlen (exportname) + 13 * n;
    s = malloc (maxlen + 1);
    if (!s) {
	PERR ("no memory to expand filename %s: %m\n", exportname);
	return NULL;
    }

    strcpy (s, exportname);
    sprintf (s_caddr, "%ld.%ld.%ld.%ld",
	     255 & (caddr >> 24), 255 & (caddr >> 16),
	     255 & (caddr >> 8), 255 & (caddr >> 0)
     );
    if (strgsub (s, "%s", s_caddr, maxlen + 1) >= 0) {
        return s;
    }
    return NULL;
  }

  /*
   * check the resources properties and set our flags accordingly.
   * Set SYNC if it's a floppy. Set  RO if we don't have write perms
   * (or if it's a cdrom?).
   *
   * and now, do file name expansion if it's necessary!
   */
  static void
  statfile (struct nbd_file *self) {

    int i;
    int errs = 0;
    char *s;

    for (i = 0; i < self->nfile; i++) {

        // PTB look for instances of %s in the resource name
        char * p = strstr(self->names[i], "%s");

        if (p != NULL) {

            if ((s = expandname (self->names[i], *self->caddr)) != NULL) {
                self->names[i] = s;
            } else {
                errs++;
                // PTB can't test if we don't know its name
                continue;
            }
        }

        if (statfile1(self->names[i], &self->flags) < 0) {
            errs++;
            continue;
        }

    }

  }

  // PTB forward decl
  static int
  getblksize (struct nbd_file *self, int *res);

  /*
   * try to open the nonopen parts of te device. Set F_OPEN when done.
   * Unset F_CLOSED if partially done. Return 0 for success, otherwise
   * something negative.
   */
  static int
  openfile(struct nbd_file * self) {

       int i;
       int errs = 0;
       int succ = 0;
       int openflags = 0;

       DEBUG("entered\n");
       if (self->flags & F_OPEN) {
           DEBUG("already open, exited %d\n", 0);
           return 0;
       }
       self->flags |= F_OPENWANTED;    // not fully open
       self->flags &= ~F_CLOSEWANTED;  // not closed


       // PTB first time through, check the file types and perms
       
       if (! (self->flags & F_STAT)) {
             // set ro/sync flags if we think we should
             unsigned long flags = self->flags;
             statfile(self);
             if (flags != self->flags) {
                 MSG("file flags changed from %#lx to %#lx\n", flags, self->flags);
             }
             // PTB we scanned once, don't do it again!
             self->flags |= F_STAT;
       }

       // now determine how we want to open
       openflags |= (self->flags&F_READONLY)? O_RDONLY : O_RDWR;
       if (self->flags &F_SYNC)
           openflags |= O_SYNC;

       #ifdef O_LARGEFILE
           openflags |= O_LARGEFILE;
       #endif

       #ifdef O_DIRECT
       if (self->flags & F_DIRECT)
           openflags |= O_DIRECT;
       #endif

       for (i = 0; i < self->nfile; i++) {

         char * exportname = self->names[i];
         int export = self->exports[i];

         if (export >= 0) {
             DEBUG("Skipped already open %s fd %d\n", exportname, export);
             succ++;
             continue;
         }

         export = open(exportname, openflags);

         if (export < 0) {
            DEBUG("Could not open exported file %s: %m\n", exportname);
            errs++;
            continue;
         }

         self->exports[i] = export;
         DEBUG("opened %s fd %d\n", exportname, export);
         // PTB checking
         if (1) {
           char c[8192], *d = c;
           int chklen;

           if (self->blksize > 0) {
               chklen = self->blksize;
           } else {
               chklen = 4096;
           }
           // PTB align at 4096 for raw devices
           d = (char*)((((unsigned long)d + 4095) >> 12) << 12);
           if (read(export, d, chklen) < chklen) {
             DEBUG("Could not read sector 0-%d of file %s fd %d: %m\n",
               ((chklen + 511) >> 9) - 1, exportname, export);
             close(export);
             errs++;
             continue;
           }
           DEBUG("Successfully read sector 0-%d of file %s fd %d\n",
               ((chklen + 511) >> 9) - 1, exportname, export);
         }

         // PTB success
         seek(export,0,SEEK_SET);
         self->lastpt[i] = 0;
         succ++;
       }
 
       self->lastpoint = 0;
       if (self->blksize <= 0) {
          getblksize(self, &self->blksize);    // PTB changes SEEK point
       }
       if (self->size <= 0) {
          getsize(self, &self->size);    // PTB changes SEEK point
          // don't know if that succeeded. Doesn't matter if not.
       }

       if (errs <= 0) {
         self->flags |= F_OPEN;          // fully open
         self->flags &= ~F_OPENWANTED;   // not fully closed
         self->lives++;                  // generation count
         DEBUG("opened file %0x for %dth time\n", (unsigned)self, self->lives);
       } 

       if (succ > 0) {
         self->flags &= ~F_CLOSED;       // not fully closed
       } 

       DEBUG("exited %d\n", -errs);
       return -errs;
  }

/*
 * only open if not open and retry set
 */
   static int
   maybeopen (struct nbd_file *self) {

       if (self->flags & F_OPEN)
	   return 0;

//       if (!(self->flags & F_OPENWANTED)) {
//	   DEBUG ("open not wanted\n");
//	   return -ENODEV;
//       }
       return self->open (self);
   }

   static int
   getblksize (struct nbd_file *self, int *res)
   {
       int i;
       int blksize = 0;

       for (i = 0; i < self->nfile; i++) {
	   int export = self->exports[i];
	   struct stat stat_info;
	   int err;
	   if (export < 0) {
	       return -ENODEV;
	   }
	   err = fstat (export, &stat_info);
	   if (err < 0) {
	       return -ENODEV;
	   }
	   if (!blksize || stat_info.st_blksize < blksize) {
	       blksize = stat_info.st_blksize;
	   }
       }
       *res = blksize;
       return 0;
   }

  /*
   * Send ioctl to ALL components of file object. Silly! Probably we
   * should refuse if more than one component, or somehow just choose
   * one. FIXME
   */
  static int ioctlfile(struct nbd_file *self, int nioctl, char *arg) {

       int i, errs = 0, res, succ = 0;

       if (maybeopen(self) < 0) {
          self->flags |= F_OPENWANTED;
          PERR("could not open target\n");
          return -ENODEV;
       }

       switch (nioctl) {
           case BLKBSZGET:
               if (self->blksize) {
                   *(int *)arg = self->blksize;
                   errs = 0;
                   goto out;
               }
               if (getblksize(self, &self->blksize) < 0) {
                   errs = -EINVAL;
                   goto out;
               }
               *(unsigned *)arg = self->blksize;
               errs = 0;
               goto out;
           case BLKBSZSET:
               self->blksize = *(int *)arg;
               errs = 0;
               goto out;
           case BLKGETSIZE:
               if (self->size) {
                   *(unsigned long *)arg = (unsigned long)(self->size >> 9);
                   errs = 0;
                   goto out;
               }
               if (self->blksize <= 0) {
                   if (getblksize(self, &self->blksize) < 0) {
                       errs = -ENODEV;
                       goto out;
                   }
               }
               if (getsize(self, &self->size) < 0) {
                   errs = -ENODEV;
                   goto out;
               }
               *(unsigned long *)arg = (unsigned long)(self->size >> 9);
               errs = 0;
               goto out;

           case BLKGETSIZE64:
               if (self->size) {
                   *(unsigned long long *)arg = self->size;
                   errs = 0;
                   goto out;
               }
               if (self->blksize <= 0) {
                   if (getblksize(self, &self->blksize) < 0) {
                       errs = -ENODEV;
                       goto out;
                   }
               }
               if (getsize(self, &self->size) < 0) {
                   errs = -ENODEV;
                   goto out;
               }
               *(unsigned long long *)arg = self->size;
               errs = 0;
               goto out;
       }

       for (i = 0; i < self->nfile; i++) {

         char * exportname = self->names[i];
         int export = self->exports[i];

         exportname = exportname;

         if (export < 0) {
             DEBUG("Skipped closed %s fd %d\n", exportname, export);
             errs++;
             continue;
         }

         res = ioctl(export, nioctl, arg);
         if (res < 0) {
            DEBUG("Could not ioctl exported file %s: %m\n", exportname);
            errs++;
            continue;
         }
         succ++;
       } // PTB end for loop

out:
       if (self->flags & F_REMOVABLE) {
           self->close(self);
           self->flags |= F_OPENWANTED;
       }
       return -errs;
  }

  /*
   * Seek to required point in file object
   *
   */
  static s64 seekfile(struct nbd_file *self, s64 offset, int whence) {

      u64 a;

      if (!(self->flags & F_OPEN)) {
        if (!(self->flags & F_OPENWANTED))
                return -ENODEV;
        if (self->open(self) < 0) {
              return -ENODEV;
        }
      }

      switch(whence) {

        case SEEK_SET:
           a = offset;
           break;

        case SEEK_CUR:
           a = self->lastpoint;
           break;

        case SEEK_END:
           a = self->size + offset;
           break;

        default:
           return -EINVAL;
      }

      if (maybeseek(self, a, NULL) < 0) {
            // PTB close the device if we had a bad seek
        if (self->flags & F_OPEN) {
          self->close(self);
          self->flags |= F_OPENWANTED;
        }
        return -EBADF;
      }
      return a;
  }
  

#if DEBUG_DFLT > 0
  /*
   * Tell us where we are right now
   * */
  static s64
  whereami(struct nbd_file *self, int fd, s64 start, int whence) {

      int i;

      if (whence == SEEK_SET)
          return start;

      for (i = 0; i < self->nfile; i++) {
          if (self->exports[i] == fd)
              break;
      }
      if (i >= self->nfile)
          return -1LL;

      switch (whence) {
          case SEEK_CUR:
              return start + self->lastpt[i];
          case SEEK_END:
              return start + self->sizes[i];
      }

      return -1LL;
  }
#endif

 /*
  * do writes from an internal buffer which has been aligned to blksize
  */
  static int
  aligned_write (int fd, char *buf, int len, int blksize) {

    static char *mybuf;
    char *dbuf;
    int wres, rres;
    int dlen = len;
    int doff = 0;
    int res = 0;

    if (!mybuf) {
        mybuf = malloc(8192);
        if (!mybuf) {
            return -ENOMEM;
        }
    }
    dbuf = (char *) (((long) &mybuf[blksize - 1]) & ~(blksize - 1));

    while (dlen >= blksize) {
	memcpy (dbuf, buf + doff, blksize);
	wres = write (fd, dbuf, blksize);
	if (wres < 0)
	    return res;
	dlen -= wres;
	res  += wres;
	doff += wres;
	if (wres < blksize)
	    return res;
    }
    while (dlen > 0) {
	// PTB pad the buffer by reading first
	rres = read (fd, dbuf, blksize);
	if (rres < 0)
	    return res;
	// PTB move back left again
	if (seek (fd, (s64) -rres, SEEK_CUR) < 0LL)
	    return res;
	// PTB overwrite buffer start with fragment
	memcpy (dbuf, buf + doff, dlen);
	wres = write (fd, dbuf, blksize);
	if (wres < 0)
	    return res;
        if (wres <= dlen) {
            dlen -= wres;
            res  += wres;
            doff += wres;
        } else {
	    if (seek (fd, (s64) -(wres - dlen), SEEK_CUR) < 0LL)
	        return res;
            dlen  = 0;
	    res  += dlen;
            doff += dlen;
        }
    }
    return res;

  }

 /*
  * do reads to an internal buffer which has been aligned to blksize
  */
  static int
  aligned_read (int fd, char *buf, int len, int blksize) {

    static char *mybuf;
    char *dbuf;
    int rres;
    int dlen = len;
    int doff = 0;
    int res = 0;

    if (!mybuf) {
        mybuf = malloc(8192);
        if (!mybuf) {
            return -ENOMEM;
        }
    }
    dbuf = (char *) (((long) &mybuf[blksize - 1]) & ~(blksize - 1));

    while (dlen >= blksize) {
	rres = read (fd, dbuf, blksize);
	if (rres < 0)
	    return res;
        memcpy(buf + doff, dbuf, rres);
	dlen -= rres;
	res  += rres;
	doff += rres;
    }
    while (dlen > 0) {
	// PTB pad the buffer by reading first
	rres = read (fd, dbuf, blksize);
	if (rres < 0)
	    return res;
	// PTB overwrite buffer with out data fragment
	memcpy (buf + doff, dbuf, dlen);
        if (rres <= dlen) {
            res += rres;
            dlen -= rres;
            doff += rres;
        } else {
            // PTB unseek padding
            if (seek (fd, (s64) -(rres - dlen), SEEK_CUR) < 0LL)
                return res;
	    res  += dlen;
            dlen  = 0;
            doff += dlen;
        }
    }
    return res;

  }


  /*
   * Reliably send/recv data to/from file object
   *
   */
  static inline int
  fileit (struct nbd_file *self, int direction, int fd, void *buf, unsigned len) {

      int res;
      unsigned count = 0;
      int loop = 0;
      static int entries;
      s64 start = seek(fd, 0LL, SEEK_CUR), finish;
      entries++;

      while (len > 0) {

         loop++;

         switch (direction) {

         case READ: // 0
           DEBUG("*[%u@%Lu]", len, self->lastpoint);

           if ((self->flags & F_DIRECT) && ((long)buf & (self->blksize - 1))) {
               res = aligned_read(fd, buf, len, self->blksize);
           } else {
               res = read(fd, buf, len);
           }

           if (res <= 0) {
               #if DEBUG_DFLT > 0
               s64 off = whereami(self, fd, 0L, SEEK_CUR);
               #endif
               DEBUG("Read len %d from %Ld failed on fd %d: %m\n", len,off,fd);
               count = (res<0) ? res : -EINVAL; 
               goto end;
           }
           DEBUG("Read %d bytes from fd %d in readfile\n", res, fd);
           break;

         default:
         case WRITE: // 1
           DEBUG("+[%u@%Lu]", len, self->lastpoint);

           if ((self->flags & F_DIRECT) && ((long)buf & (self->blksize - 1))) {
               res = aligned_write(fd, buf, len, self->blksize);
           } else {
               res = write(fd, buf, len);
           }

           if (res <= 0) {
               #if DEBUG_DFLT > 0
               s64 off = whereami(self, fd, 0L, SEEK_CUR);
               #endif
               DEBUG("Write len %d at %Ld failed on fd %d: %m\n", len, off, fd);
               count = (res<0) ? res : -EINVAL; 
               goto end;
           }
           DEBUG("Wrote %d bytes to fd %d in readfile\n", res, fd);
           if (self->flags & F_SYNC) {
               fsync(fd);
           }
           break;
         }
         len   -= res;
         buf   += res;
         count += res;
      }

end:
      finish = seek(fd, 0LL, SEEK_CUR);
      if (count >= 0) {
          // PTB miscount
          if (count != (finish - start)) {
          PERR("counted %d but seeked over %Ld from %#Lx-%#Lx!\n",
                  count, finish - start, start, finish);
          }
      } else {
          // PTB negative values are straight errors
      }
      return count;
   }


  static int
  unlockfile(struct nbd_file * self, u64 from, s64 len) {
    int err;
    struct flock lock = {
        l_type:   F_UNLCK,
        l_whence: SEEK_END,
        l_start:  from,
        l_len:    len,
        l_pid:    self->pid,
    };
    int fd;
    struct nbd_seek s;

    // maybe we should go through that open palavar!
      
    if (maybeopen(self) < 0) {
          self->flags |= F_OPENWANTED;
          return -ENODEV;
    }

    if (resolve_offset(self, self->size - 1, &s) < 0)
          return -EINVAL;

    if ((fd = s.fd) < 0)
          return -EINVAL;
   
    err = fcntl (fd, F_SETLKW, &lock);
    DEBUG("   lock %Ld-%Ld ->\n", from, from+len);
    return err;
  }

  static int
  lockfile(struct nbd_file * self, u64 from, s64 len, int type) {

      // PTB we lock a virtual area beyond the last file component

      struct flock lock = {
        l_type: type,
        l_whence: SEEK_END,
        l_start: from,
        l_len: len,
        l_pid: self->pid,
      };
      const unsigned long udelay = 100000;
      int count = 30;
      struct nbd_seek s;
      int fd;

      // maybe we should go through that open palavar!
      
      if (maybeopen(self) < 0) {
          self->flags |= F_OPENWANTED;
          return -ENODEV;
      }

      if (resolve_offset(self, self->size - 1, &s) < 0)
          return -EINVAL;

      if ((fd = s.fd) < 0)
          return -EINVAL;
   
      // PTB spin for lock
      while (fcntl(fd, F_SETLKW, &lock) < 0) {
          int err = errno;

          switch (err) {

              case EINTR:
              case EACCES:
              case EAGAIN:
                  break;
              case EDEADLK:
                  if (--count < 0) {
                      DEBUG("resource %Ld-%Ld is already %slocked\n",
                              from, from+len, type == F_WRLCK ? "w" : "r");
                      return -err;
                  }
                  break;
              default:
                return -err;
          }

          microsleep(udelay);
          lock.l_type = type;
          lock.l_whence = SEEK_END;
          lock.l_start = from;
          lock.l_len = len;
          lock.l_pid = self->pid;
      }
      DEBUG("-> %slock %Ld-%Ld\n", type == F_WRLCK ? "w" : "r", from, from+len);
      return 0;
  }


/*
 * only close if not closed and retry set
 */
   static int
   maybeclose (struct nbd_file * self) {

       if (self->flags & F_CLOSED)
	   return 0;

       return self->close(self);
   }

   static int
   maybereopen (struct nbd_file *self) {
       maybeclose(self);
       self->flags |= F_OPENWANTED;
       return maybeopen(self);
   }

   static int opfile_linear(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len);
   static int opfile_stripe(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len);
   static int opfile_mirror(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len);

/*
 * This routine selects which component of the mirror devices to write
 * to, and how to break up the write.
 */
   static int
   opfile(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len) {

            // split into sections of at most blocksize

            int tot     = 0;  // total bytes transfered so far
            short didlock = 0;
            unsigned locklen = len;
            u64      lockoff = foffset;

            DEBUG("opfile enters for type %d len %d from %Ld\n",
                  direction, len, foffset);

            if (maybeopen(self) < 0) {
              self->flags |= F_OPENWANTED;
              PERR("could not open target\n");
              return -ENODEV;
            }

            if ( self->flags & F_DOLOCK) {
                int res;
                int l_type;
                switch(direction) {
                        case READ:
                        l_type = F_RDLCK;
                        break;

                        case WRITE:
                        l_type = F_WRLCK;
                        break;

                        default:
                        tot = -EINVAL;
                        goto fail;
                        break;
                }
                res = lockfile(self, foffset, len, l_type);
                if (res < 0)  {
                        tot = res;
                        goto fail;
                }
                didlock = 1;
            }

            switch(self->mode) {
              default:
                PERR("unknown mode %d!\n", self->mode);
                tot = -1;
              break;
              case F_LINEAR_MODE:
                tot = opfile_linear(self, direction, buf, foffset, len);
              break;
              case F_STRIPE_MODE:
                tot = opfile_stripe(self, direction, buf, foffset, len);
              break;
              case F_MIRROR_MODE:
                tot = opfile_mirror(self, direction, buf, foffset, len);
              break;
            }

            if (tot < len) {
              PERR("short file op, %d instead of %d bytes\n", tot, len);
              goto fail;
            }

            if (didlock)
                unlockfile(self, lockoff, locklen);
            DEBUG("opfile exits OK result %d for type %d len %d from %Ld\n",
                  tot, direction, locklen, lockoff);
            return tot;

         fail:

            if (didlock)
                unlockfile(self, lockoff, locklen);

            maybeclose(self);
            self->flags |= F_OPENWANTED;

            PERR("opfile exits FAIL result %d for type %d len %d from %Ld\n",
                  tot, direction, len, foffset);
            return tot;
    }

   static int
   opfile_stripe(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len) {

            int boffset = 0;  // current offset within buf
            int tot     = 0;  // total bytes transfered so far
            const int retries
                        = 2;  // number of tries at an op allowed
            int tries   = 0;  // number of tries so far

            while (len > 0) {

              int blockoffset = foffset % self->blksize; // offset within block
              unsigned sublen;                 // transfer this much
              int export = -1;
              int count  = -1;

              // PTB cut sublen down to go up to block boundary only

              if ( len + blockoffset > self->blksize )
                sublen = self->blksize - blockoffset;
              else
                sublen = len;

           try_seek:

              // PTB this sets self->lasti, self->lastpoint, self->lastpt[lasti]
              export   = maybeseek(self, foffset, NULL);

              //seek (export, self->lastpt[self->lasti], SEEK_SET);
              DEBUG("found offset %Ld in fd %d\n", foffset, export);

              // PTB returning INVAL this early is problematic, since we
              // may need to return "tot = 0" to properly indicate the
              // failure  type (local device).

              // At this point I can't tell whether the
              // instruction is wrong (overlong, underzero) or the
              // device really failed.
              //  Maybe it needs maybeseek to return
              // the device index and relative offset, but not do the
              // seek, and we do the seek here?

              if (export < 0) {
                if (tries++ < retries) {
                  if (maybereopen(self) >= 0) {
                      PERR("seek failed at %Ld! retrying\n", foffset);
                    goto try_seek;
                  }
                }
                tot = -EINVAL;
                PERR("seek failed at %Ld!\n", foffset);
                goto fail;
              }

            try_fileit:

              count = fileit(self, direction, export, buf + boffset, sublen);

              if (count <= 0) {
                if (tries++ < retries) 
                  goto try_fileit;
                goto fail;
              }

              self->lastpoint += count;
              self->changedfd = 0;
              self->lastpt[self->lasti] += count;

              // PTB check if we can use the same seek point next time

              if (count + blockoffset >= self->blksize) {
                  self->changedfd = 1;
              }

              foffset += count;
              boffset += count;
              len     -= count;
              tot     += count;
            }

            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, locklen, lockoff);
            return tot;

         fail:

            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, len, foffset);
            return tot;
    }


   static int
   opfile_mirror(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len) {

            int boffset = 0;  // current offset within buf
            int tot     = 0;  // total bytes transfered so far
            const int retries
                        = 2;  // number of tries at an op allowed
            int tries   = 0;  // number of tries so far

            while (len > 0) {

              int blockoffset = foffset % self->blksize; // offset within block
              unsigned sublen;                 // transfer this much
              int export = -1;
              int count  = -1;

            
              // PTB cut sublen down to go up to block boundary only

              if ( len + blockoffset > self->blksize )
                sublen = self->blksize - blockoffset;
              else
                sublen = len;

           try_seek:

              // PTB this sets self->lasti, self->lastpoint, self->lastpt[lasti]
              export   = maybeseek(self, foffset, NULL);


              //seek (export, self->lastpt[self->lasti], SEEK_SET);
              DEBUG("found offset %Ld in fd %d\n", foffset, export);

              // PTB returning INVAL this early is problematic, since we
              // may need to return "tot = 0" to properly indicate the
              // failure  type (local device).

              // At this point I can't tell whether the
              // instruction is wrong (overlong, underzero) or the
              // device really failed.
              //  Maybe it needs maybeseek to return
              // the device index and relative offset, but not do the
              // seek, and we do the seek here?

              if (export < 0) {
                if (tries++ < retries) {
                  if (maybereopen(self) >= 0) {
                      PERR("retry seek\n");
                    goto try_seek;
                  }
                }
                tot = -EINVAL;
                goto fail;
              }

            try_fileit:

                switch(direction) {
                int j; 
                fd_set fds, * rfds;
                int fmax;

                case READ: // 0

                  rfds = & fds; 
                  fmax = 0;
                  FD_ZERO(rfds);
                  for (j = 0; j < self->nfile; j++) {
                    int fd = self->exports[j];
                    fmax = fmax < fd ? fd : fmax;
                    FD_SET(fd, rfds);
                  }

                  if (select(fmax+1,rfds,NULL,NULL,NULL) <= 0)
                    break;

                  for (j = 0; j < self->nfile; j++) {
                    int fd = self->exports[j];
                    if (FD_ISSET(fd, rfds)){
                      export = fd;
                      break;
                    }
                  }

                  count = fileit(self, direction, export, buf + boffset, sublen);
                  break;

                case WRITE: // 1
                  for (j = 0; j < self->nfile; j++) {
                    count = fileit(self, direction, self->exports[j], buf+boffset, sublen);
                    if (count <= 0)
                      break;
                  }
                  break;
                }

              if (count <= 0) {
                if (tries++ < retries) 
                  goto try_fileit;
                goto fail;
              }

              self->lastpoint += count;
              self->changedfd = 0;
              self->lastpt[self->lasti] += count;

              // PTB check if we can use the same seek point next time


              foffset += count;
              boffset += count;
              len     -= count;
              tot     += count;
            }

            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, len, foffset0);
            return tot;

         fail:

            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, len, foffset0);
            return tot;
    }


   static int
   opfile_linear(struct nbd_file * self, int direction, char * buf, u64 foffset, unsigned len) {

            int boffset = 0;  // current offset within buf
            int tot     = 0;  // total bytes transfered so far
            const int retries
                        = 2;  // number of tries at an op allowed
            int tries   = 0;  // number of tries so far

            int loops = 0;
            static int entries = -1;

            DEBUG("entry %#Lx-%#Lx direction %s\n", foffset, foffset+len,
                    direction == READ? "read" : "write");

            entries++;

            while (len > 0) {

              int blockoffset = foffset % self->blksize; // offset within block
              unsigned sublen;                 // transfer this much
              int export = -1;
              int count  = -1;
              struct nbd_seek s;

              loops++;

            
              // PTB cut sublen down to go up to block boundary only

              //if ( len + blockoffset > self->blksize )
              //  sublen = self->blksize - blockoffset;
              //else
              // sublen = len;


              if (resolve_offset(self, foffset, &s) >= 0) {
                s64 localstart  = foffset - s.a;
                s64 localfinish = localstart + self->sizes[s.i];
                s64 locallen    = localfinish - s.a;
                if (len > locallen) {
                    sublen = locallen;
                } else {
                    sublen = len;
                }
                s.direction = 0; // is input for following maybeseek
              } else {
                // cut to blocklength - we still don't know where we are
                if ( len + blockoffset > self->blksize )
                  sublen = self->blksize - blockoffset;
                else
                  sublen = len;
                s.direction = 1; // is output for following maybeseek
              }

           try_seek:

              // PTB this sets self->lasti, self->lastpoint, self->lastpt[lasti]
              export   = maybeseek(self, foffset, &s);

              //seek (export, self->lastpt[self->lasti], SEEK_SET);
              DEBUG("found offset %Ld in fd %d\n", foffset, export);

              // PTB returning INVAL this early is problematic, since we
              // may need to return "tot = 0" to properly indicate the
              // failure  type (local device).

              // At this point I can't tell whether the
              // instruction is wrong (overlong, underzero) or the
              // device really failed.
              //  Maybe it needs maybeseek to return
              // the device index and relative offset, but not do the
              // seek, and we do the seek here?

              if (export < 0) {
                if (tries++ < retries) {
                  if (maybereopen(self) >= 0) {
                      PERR("retry seek\n");
                    goto try_seek;
                  }
                }
                tot = -EINVAL;
                goto fail;
              }

            try_fileit:


              count = fileit(self, direction, export, buf + boffset, sublen);


              if (count <= 0) {
                if (tries++ < retries) 
                  goto try_fileit;
                goto fail;
              }

              self->lastpoint += count;
              self->changedfd = 0;
              self->lastpt[self->lasti] += count;

              // PTB check if we can use the same seek point next time

              if (self->lastpt[self->lasti] >= self->sizes[self->lasti]) {
                  self->changedfd = 1;
              }

              foffset += count;
              boffset += count;
              len     -= count;
              tot     += count;
            }

            if (self->flags & F_REMOVABLE) {
                self->close(self);
                self->flags |= F_OPENWANTED;
            }
            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, len0, lockoff);
            return tot;

         fail:

            if (self->flags & F_REMOVABLE) {
                self->close(self);
                self->flags |= F_OPENWANTED;
            }
            DEBUG("opfile exits with result %d for type %d len %d from %Ld\n",
                  tot, direction, len0, foffset);
            return tot;
    }

   static int
   readfile(struct nbd_file * self, char * buf, unsigned len, u64 foffset) {
           int res = opfile(self, 0,buf,foffset,len);
           DEBUG("readfile totalled %d bytes from %Ld\n", res, foffset);
           return res;
   }

   static int
   writefile(struct nbd_file * self, char * buf, unsigned len, u64 foffset) {
           int res = opfile(self, 1,buf,foffset,len);
           DEBUG("writefile totalled %d bytes from %Ld\n", res, foffset);
           return res;
   }

   static int
   syncfile(struct nbd_file * self) {
           int i, err = 0;
           for (i = 0; i < self->nfile; i++) {
             if (self->exports[i] >=0)
               err += fsync(self->exports[i]);
           }
           return err;
   }


   static int
   reopenfile(struct nbd_file * self) {

         int errs;

         if (! (self->flags & F_CLOSED)) {
           errs = self->close(self);
           if (errs < 0)
             return errs;
         }

         errs = self->open(self);

         return errs;
   }


  int init_file(struct nbd_file * self, int nfile, int mode, s64 *sizes,
          int blksize, char* names[], int ro, int sync, int async, int lock,
          int direct, int removable, unsigned long *caddr) {

       int i;

       self->nfile = nfile;
       self->mode  = mode ;
       self->caddr = caddr; // this is an alias back to the real caddr

       self->size  = 0;
       self->flags = 0;

       self->lastpoint = 0;
       self->changedfd = 0;

       self->lives = 0;

       for (i = 0; i < nfile; i++) {
           self->names[i]     = names[i];
           self->exports[i]   = -1;
           self->sizes[i]     = -1;
       }

       if (ro) 
         self->flags |= F_READONLY;
       if (sync && ! async)
         self->flags |= F_SYNC;
       if (async && ! sync)
         self->flags |= F_ASYNC;
       if (lock) 
         self->flags |= F_DOLOCK;
       if (direct) 
         self->flags |= F_DIRECT;
       if (removable) 
         self->flags |= F_REMOVABLE;

       self->write = writefile;
       self->read  = readfile;
       self->sync  = syncfile;
       self->reopen= reopenfile;
       self->open  = openfile;
       self->seek  = seekfile;
       self->ioctl = ioctlfile;
       self->close = closefile; // PTB !!! may cause problems 
       self->lock  = lockfile;
       self->unlock= unlockfile;
       self->flags |= F_CLOSED;

       maybeopen(self);
       // PTB maybe test for blksize
       if (blksize <= 0) {
           if (getblksize(self, &self->blksize) < 0) {
               self->blksize = 0;
           }
       } else {
           self->blksize = blksize;
       }

       if (self->blksize > 0) {
           // PTB if one component size was not given, then we don't know total
           if (getsize(self, &self->size) < 0) {
               self->size = 0;
           } 
       } else {
           self->size = 0;
       }
       maybeclose(self);


       return 0;
  }

